@@ -7,21 +7,55 @@ use Carp;
 use Scalar::Util 'blessed';
 use MRO::Compat;
 
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 
 our @EXPORT = qw(before around after);
+our @EXPORT_OK = qw(guard);
+our %EXPORT_TAGS = (all => [@EXPORT, @EXPORT_OK], guard => [@EXPORT, 'guard']);
+
+################################################################################
+
+# if you're interested in doing very dynamic things with this module (I
+# certainly am) then there are some undocumented tricks. most of these
+# should become public interface.
+
+# Class::Method::Modifiers::_wipeout
+# will clear the modifier stash of:
+#     (no argument):   everything
+#     package name:    all methods in that package
+#     package::method: this one method in package
+# this is particularly useful in conjunction with Module::Refresh. without
+# _wipeout, refreshing a module with modifiers will cause the method to be
+# wrapped by each modifier twice
+
+# null modifier
+# passing "-" to the _install method (which is what the before/after/around
+# functions thinly wrap) will let you install a null modifier. this is most
+# useful if you want to change the original method that is being wrapped
+# without affecting the wrappers at all
+
+# see also Calf, which is a basic Moose clone optimized for highly pluggable,
+# dynamic apps
+
+################################################################################
 
 # this is a dynamically scoped variable that has, during an around, the arounds
 # that must be called "inside" that around. this includes the original method.
 # implemented so multiple arounds by the same class does the right thing
 our @AROUNDS_LEFT;
 
-# keeps track of what modifiers we've defined
+# method modifier symbol table. keys are of the form Npackage::method where N
+# is BAC for the three modifiers, O for the original sub we needed to
+# overwrite, X for dispatch cache, and - for null modifiers
+# I've considered making this public, but keeping all the magic in one place
+# is the best way to go
 my %method_cache;
 
-my %type_expand = (A => "after", B => "before", C => "around");
+my %type_expand = (A => "after", B => "before", C => "around", G => "guard");
 
-# this is the method that gets injected into each class
+# this is the method that gets injected into each method slot, used to handle
+# the calling of modifiers and eventually up to a parent class, or the original
+# sub that was there before we overwrote it
 sub _resolve
 {
     my $methodname = shift;
@@ -51,15 +85,21 @@ sub _resolve
         $method_cache{"X$qualified"} = $dispatch;
     }
 
-    my $before = $method_cache{"B$qualified"} || [];
-    my $after  = $method_cache{"A$qualified"} || [];
-    my $around = $method_cache{"C$qualified"} || [];
+    my $before  = $method_cache{"B$qualified"} || [];
+    my $after   = $method_cache{"A$qualified"} || [];
+    my $around  = $method_cache{"C$qualified"} || [];
+    my $guard   = $method_cache{"G$qualified"} || [];
 
     for (@$before)
     {
         $_->(@_);
     }
 
+    for (@$guard)
+    {
+        $_->(@_) or return;
+    }
+
     my @ret;
 
     {
@@ -82,7 +122,7 @@ sub _resolve
     return wantarray ? @ret : $ret[0];
 }
 
-# this handles the injection and error checking
+# this handles the injection. thinly wrapped by before/after/around
 sub _install
 {
     my $mod_type   = shift;
@@ -100,14 +140,19 @@ sub _install
         $already_installed = 1;
 
         # if we have an existing method, and we don't know about it, that is
-        # a "sub foo" that we probably don't want
-        if (!exists($method_cache{"A$qualified"})
+        # a "sub foo" that we must cache and overwrite. the special modifier
+        # "-" (which is the null modifier) may be used to bypass the check
+        # for existing modifiers
+        if ($mod_type eq '-' ||
+           (!exists($method_cache{"A$qualified"})
          && !exists($method_cache{"B$qualified"})
-         && !exists($method_cache{"C$qualified"}))
+         && !exists($method_cache{"C$qualified"})
+         && !exists($method_cache{"G$qualified"})
+         && !exists($method_cache{"-$qualified"})))
         {
-            # this code path is hit when you say 'sub foo' 'around foo =>'
             $method_cache{"O$qualified"} = *{$qualified}{CODE};
             $already_installed = 0;
+            delete $method_cache{"X$qualified"}; # clear dispatch cache
         }
     }
 
@@ -124,24 +169,25 @@ sub _install
 
     *{$qualified} = sub
     {
+        # this helps _resolve figure out what's going on without relying on
+        # caller
         unshift @_, $methodname, $package;
+
         goto &_resolve;
     } unless $already_installed;
 }
 
 # this is used to clear any method modifiers in the internal cache
-# used in conjunction with Module::Refresh, I hear..
 sub _wipeout
 {
-    my $self = shift;
-
     if (!@_)
     {
         for (values %method_cache) { undef $_ }
         return;
     }
 
-    my $package = shift(@_) . '::';
+    my $package = shift(@_);
+    $package = (blessed $package || $package) . '::';
     my $method = $package . shift(@_) if @_;
 
     for my $key (keys %method_cache)
@@ -149,12 +195,12 @@ sub _wipeout
         my $k = substr($key, 1); # get rid of the modifier type
         if ($method)
         {
-            undef $method_cache{$key}
+            do { warn $key; undef $method_cache{$key} }
                 if $k eq $method;
         }
         else
         {
-            undef $method_cache{$key}
+            do { warn $key; undef $method_cache{$key} }
                 if substr($k, 0, length($package)) eq $package;
         }
     }
@@ -169,7 +215,8 @@ sub _orig
     my $next = shift @AROUNDS_LEFT
         or die "It looks like you're calling \$orig more than once in around. Don't!!";
 
-    # need to set up the next $orig
+    # need to set up the next $orig, except when we're dispatching next to
+    # the parent class or the original method in the class
     unshift @_, \&_orig if @AROUNDS_LEFT;
 
     goto &$next;
@@ -202,13 +249,22 @@ sub around
     }
 }
 
+sub guard
+{
+    my $modifier = pop;
+    for (@_)
+    {
+        _install('G', $_, $modifier);
+    }
+}
+
 =head1 NAME
 
 Class::Method::Modifiers - provides Moose-like method modifiers
 
 =head1 VERSION
 
-Version 0.06 released 10 Sep 07
+Version 0.07 released 12 Sep 07
 
 =head1 SYNOPSIS
 
@@ -239,12 +295,13 @@ In its most basic form, a method modifier is just a method that calls
 C<< $self->SUPER::foo(@_) >>. I for one have trouble remembering that exact
 invocation, so my classes seldom re-dispatch to their base classes. Very bad!
 
-C<Class::Method::Modifiers> provides three modifiers: C<before>, C<around>,
-and C<after>. C<before> and C<after> are run just before and after the method
-they modify, but can not really affect that original method. C<around> is run
-in place of the original method, with a hook to easily call that original
-method. See the C<MODIFIERS> section for more details on how the particular
-modifiers work.
+C<Class::Method::Modifiers> provides four modifiers: C<before>, C<around>,
+C<after>, and C<guard>. C<before> and C<after> are run just before and after
+the method they modify, but can not really affect that original method.
+C<guard> is much like C<before>, except that it can prevent the execution of
+the original method. C<around> is run in place of the original method, with a
+hook to easily call that original method. See the C<MODIFIERS> section for more
+details on how the particular modifiers work.
 
 One clear benefit of using C<Class::Method::Modifiers> is that you can define
 multiple modifiers in a single namespace. These separate modifiers don't need
@@ -254,7 +311,7 @@ those methods to flesh out the specifics.
 
 Parent classes need not know about C<Class::Method::Modifiers>. This means you
 should be able to modify methods in I<any> subclass. See
-L<Term::VT102::OneBased> for an example of subclassing with CMM.
+L<Term::VT102::ZeroBased> for an example of subclassing with CMM.
 
 In short, C<Class::Method::Modifiers> solves the problem of making sure you
 call C<< $self->SUPER::foo(@_) >>, and provides a cleaner interface for it.
@@ -269,6 +326,23 @@ would have received. You can modify the C<@_> the original method will receive
 by changing C<$_[0]> and friends (or by changing anything inside a reference).
 This is a feature!
 
+=head2 guard method(s) => sub { ... }
+
+C<guard> is called before the method it is modifying, between any C<before>s
+and any C<around>s. If the guard returns a true value, then execution will
+proceed as normal. If the guard returns a false value, then any further
+C<guard>s, C<around>s, C<after>s are not run. The method will appear to have
+returned the canonical false value (C<undef> in scalar context, the empty list
+in list context). It receives the same C<@_> as the the method it is modifying
+would have received. You can modify the C<@_> the original method will receive
+by changing C<$_[0]> and friends (or by changing anything inside a reference).
+This is a feature!
+
+Since C<guard> is not exported by default, you must import either 'guard',
+':guard', or ':all' when C<use>-ing CMM. 'guard' will import only this one
+modifier, ':guard' will import guard and before/after/around, and ':all' will
+import guard, before, after, around, and anything else added in the future.
+
 =head2 after method(s) => sub { ... }
 
 C<after> is called after the method it is modifying. Its return value is
@@ -317,11 +391,12 @@ You can use C<around> to:
 
 =head1 NOTES
 
-All three modifiers; C<before>, C<after>, and C<around>; are exported into your
-namespace by default. You may C<use Class::Method::Modifiers ()> to avoid
-thrashing your namespace. I may steal more features from L<Moose>, namely
-C<super>, C<override>, C<inner>, C<augment>, and whatever the L<Moose> folks
-come up with next.
+All three normal modifiers; C<before>, C<after>, and C<around>; are exported
+into your namespace by default. C<guard> is imported only if you ask for it
+(such as by C<use Class::Method::Modifiers ':guard'> or C<':all'>. You may
+C<use Class::Method::Modifiers ()> to avoid thrashing your namespace. I may
+steal more features from L<Moose>, namely C<super>, C<override>, C<inner>,
+C<augment>, and whatever the L<Moose> folks come up with next.
 
 Note that the syntax and semantics for these modifiers is directly borrowed
 from L<Moose> (the implementations, however, are not).
